/***********************************************************
 * $Id: AbstractResourceSerlvlet.java 880 2013-07-26 09:52:19Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 26.07.2013
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>An abstract base servlet, which provides the additional header functionality
 * for concrete implementations.</p>
 * 
 * <p>Implementation od his servlet may be configured through initialization parameters
 * inside a classical servlet container like in the following <code>web.xml</code>
 * example:</p>
 * <pre>
 * &lt;servlet&gt;
 *  ...
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.0&lt;/param-name&gt;
 *   &lt;param-value&gt;X-Frame-Options: DENY&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.0.mimeTypeRegex&lt;/param-name&gt;
 *   &lt;param-value&gt;text/html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.1&lt;/param-name&gt;
 *   &lt;param-value&gt;X-Frame-Options: SAMEORIGIN&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.1.pathRegex&lt;/param-name&gt;
 *   &lt;param-value&gt;.*\.cache\.html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.2&lt;/param-name&gt;
 *   &lt;param-value&gt;X-UA-Compatible: IE=edge&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.2.mimeTypeRegex&lt;/param-name&gt;
 *   &lt;param-value&gt;text/html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 * &lt;/servlet&gt;
 * </pre>
 * 
 * <p>In OSGi frameworks it is recommended to set up implementation od this servlet using
 * blueprint or a similar technology using the supplied bean property setters like in the
 * following example:</p>
 * <pre>
 *  &lt;bean id="resourceServlet" class="...implementation of AbstractResourceServlet..."&gt;
 *    ...
 *    &lt;bp:property name="additionalHeaders"&gt;
 *      &lt;bp:list&gt;
 *        &lt;bp:bean class="org.clazzes.util.http.AdditionalHeader"&gt;
 *          &lt;bp:property name="header" value="X-Frame-Options"/&gt;
 *          &lt;bp:property name="value" value="DENY"/&gt;
 *          &lt;bp:property name="mimeTypeRegex" value="text/html"/&gt;
 *        &lt;/bp:bean&gt;
 *        &lt;bp:bean class="org.clazzes.util.http.AdditionalHeader"&gt;
 *          &lt;bp:property name="header" value="X-Frame-Options"/&gt;
 *          &lt;bp:property name="value" value="SAMEORIGIN"/&gt;
 *          &lt;bp:property name="pathRegex" value=".*\.cache\.html"/&gt;
 *        &lt;/bp:bean&gt;
 *        &lt;bp:bean class="org.clazzes.util.http.AdditionalHeader"&gt;
 *          &lt;bp:property name="header" value="X-UA-Compatible"/&gt;
 *          &lt;bp:property name="value" value="IE=edge"/&gt;
 *          &lt;bp:property name="mimeTypeRegex" value="text/html"/&gt;
 *        &lt;/bp:bean&gt;
 *      &lt;/bp:list&gt;
 *    &lt;/bp:property&gt;
 *  &lt;/bean&gt;
 * </pre>
 */
public class AbstractResourceSerlvlet extends HttpServlet {

    private static final long serialVersionUID = -4240361333446135052L;

    private static final Logger log = LoggerFactory
            .getLogger(AbstractResourceSerlvlet.class);

    private List<AdditionalHeader> additionalHeaders;

    /**
     * Set matching additional headers on the given servlet response based on
     * the given resolved path and MIME tpye.
     * 
     * @param resp The servlet response to set matching additional headers on.
     * @param resolvedPath
     *            The resolved path starting with a slash or an empty string, if
     *            there was no path component after the serlvet's own path.
     * @param mimeType
     *            The MIME type of the delivered resource or <code>null</code>,
     *            if no MIME type match should be performed.
     */
    protected void propagateAdditionalHeaders(HttpServletResponse resp,
            String resolvedPath, String mimeType) {

        if (this.additionalHeaders != null) {

            for (AdditionalHeader ah : this.additionalHeaders) {

                if (ah.getPathRegex() != null) {

                    if (!ah.getPathRegex().matcher(resolvedPath).matches()) {
                        continue;
                    }
                }

                if (ah.getMimeTypeRegex() != null) {

                    if (mimeType == null
                            || !ah.getMimeTypeRegex().matcher(mimeType)
                                    .matches()) {
                        continue;
                    }
                }

                if (log.isDebugEnabled())
                    log.debug("Sending additional header [{}: {}].",
                            ah.getHeader(), ah.getValue());

                resp.setHeader(ah.getHeader(), ah.getValue());
            }
        }
    }

    /**
     * @param additionalHeaders
     *            Additional headers to be sent based on path and/or MIME type
     *            regex matching.
     */
    public void setAdditionalHeaders(List<AdditionalHeader> additionalHeaders) {
        this.additionalHeaders = additionalHeaders;
    }

    /**
     * @param additionalHeader
     *            An additional header to send.
     */
    public void addAdditionalHeader(AdditionalHeader additionalHeader) {

        if (this.additionalHeaders == null) {
            this.additionalHeaders = new ArrayList<AdditionalHeader>();
        }

        this.additionalHeaders.add(additionalHeader);
    }

    /**
     * @return Additional headers to be sent.
     */
    public List<AdditionalHeader> getAdditionalHeaders() {
        return this.additionalHeaders;
    }

    
    
    
    @Override
    public void init() throws ServletException {

        ServletConfig cfg = this.getServletConfig();
        
        int i = 0;
        
        while(true) {
        
            String cfgKey = "additionalHeader."+i;
            
            String p = cfg.getInitParameter(cfgKey);
                        
            if (p == null) {
                break;
            }
            
            String[] kv = p.split(":",2);
            
            if (kv.length != 2) {
                
                throw new ServletException("The value ["+p+"] of configuration key ["+cfgKey+"] does not contain a colon to separate the name from the value of an additional header.");
            }
            
            AdditionalHeader ah = new AdditionalHeader();
            ah.setHeader(kv[0].trim());
            ah.setValue(kv[1].trim());
            
            p = cfg.getInitParameter(cfgKey+ ".pathRegex");
            
            if (p != null) {
                
                ah.setPathRegex(Pattern.compile(p));
            }
            
            p = cfg.getInitParameter(cfgKey+ ".mimeTypeRegex");
            
            if (p != null) {
                
                ah.setMimeTypeRegex(Pattern.compile(p));
            }

            this.addAdditionalHeader(ah);
            
            ++i;
        }
    }
}
