/***********************************************************
 * $Id: AcceptHelper.java 505 2011-06-16 10:06:40Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 16.06.2011
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Provide selections of items from <code>Accept</code>, <code>Accept-Charset</code>,
 * <code>Accept-Encoding</code> or <code>Accept-Language</code> HTTP headers.
 */
public abstract class AcceptHelper {

	private static final Logger log = LoggerFactory.getLogger(AcceptHelper.class);

    /**
     * Evaluate an <code>Accept-*</code> header and match it against a list of supported
     * server-side items based on exact string matching.
     * 
     * @param acceptedItems The value of an <code>Accept-*</code> header.
     * @param supportedItems The list of server-side items. (locales, encodings,...)
     * @return The sever-side item, which best matches the client's needs or
     *         <code>null</code>, if no requested item matches any of the server-side
     *         values.
     */
    public static String selectAcceptItem(String acceptedItems, String[] supportedItems) {

    	return selectAcceptItem(acceptedItems,supportedItems,null);
    }
    
    /**
     * Evaluate an <code>Accept-*</code> header and match it against a list of supported
     * server-side items.
     * 
     * @param acceptedItems The value of an <code>Accept-*</code> header.
     * @param supportedItems The list of server-side items. (locales, encodings,...)
     * @param matcher A matcher for allowing fuzzy matches or <code>null</code> to
     *                use exact matching through {@link String#equals(Object)}.
     * @return The sever-side item, which best matches the client's needs or
     *         <code>null</code>, if no requested item matches any of the server-side
     *         values.
     */
    public static String selectAcceptItem(String acceptedItems, String[] supportedItems, AcceptMatcher matcher) {
        
        String selectedItem = null;     
        
        if (acceptedItems != null && supportedItems != null && supportedItems.length > 0) {
            
            String[] encodingsWithQ = acceptedItems.split(",");
            double qSelected = -1.0;
            
            for (int i=0;i<encodingsWithQ.length;++i) {
                
                try {
                    String[] encq = encodingsWithQ[i].split(";",2);
                    
                    double q;
                    
                    if (encq.length > 1) {
                        
                        encq[1] = encq[1].trim();
                        if (encq[1].startsWith("q=")) 
                            q = Double.parseDouble(encq[1].substring(2));
                        else
                            throw new IllegalArgumentException("No valid q-value after semicolon.");
                    }
                    else
                        q = 1.0;
                    
                    if (q > qSelected) {
                        
                        encq[0] = encq[0].trim();
                
                        if ("*".equals(encq[0])) {
                            selectedItem = supportedItems[0];
                            qSelected = q;
                        }
                        else {
                            
                           	if (matcher == null) {
                           		
                           		for (String enc : supportedItems) {
                           			
                           			if (encq[0].equals(enc)) {
                           				selectedItem = enc;
                           				qSelected = q;
                           				break;
                           			}
                           		}
                          	}
                           	else {
                           		for (String enc : supportedItems) {
                           			
                           			double qm = q * matcher.matchItem(encq[0],enc);
                           			
                           			if (qm > qSelected) {
                           				selectedItem = enc;
                           				qSelected = qm;
                           			}
                           		}
                           	}
                        }
                    }
                } catch (Throwable e) {
                    log.error("Received invalid accept part ["+ encodingsWithQ[i]+"]: "+e.getMessage());
                }
            }
            if (qSelected <= 0.0)
                selectedItem = null;
        }
        return selectedItem;
    }

}
