/***********************************************************
 * $Id: ResourceServlet.java 907 2013-10-11 18:53:00Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 17.09.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.regex.Pattern;

import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.codec.binary.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>A servlet, which exports parts of the {@link ServletContext}'s resources as HTTP resources
 * with configuration options on which items should be delivered in compressed form.</p>
 * 
 * <p>This servlet may be configured through initialization parameters inside a classical
 * servlet container like in the following <code>web.xml</code> example:
 * </p>
 * <pre>
 * &lt;servlet&gt;
 *  &lt;servlet-name&gt;foobar-resources&lt;/servlet-name&gt;
 *  &lt;servlet-class&gt;org.clazzes.util.http.ResourceServlet&lt;/servlet-class&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;resourcePath&lt;/param-name&gt;
 *   &lt;param-value&gt;/foobar&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;excludeMimeType&lt;/param-name&gt;
 *   &lt;param-value&gt;image/png,image/jpeg,image/gif&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;aliases&lt;/param-name&gt;
 *   &lt;param-value&gt;/=/index.html,/index.htm=/index.html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;redirects&lt;/param-name&gt;
 *   &lt;param-value&gt;/app1=appServlet,/app2=/ctx2/appServlet&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.0&lt;/param-name&gt;
 *   &lt;param-value&gt;X-Frame-Options: DENY&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.0.mimeTypeRegex&lt;/param-name&gt;
 *   &lt;param-value&gt;text/html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.1&lt;/param-name&gt;
 *   &lt;param-value&gt;X-Frame-Options: SAMEORIGIN&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.1.pathRegex&lt;/param-name&gt;
 *   &lt;param-value&gt;.*\.cache\.html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.2&lt;/param-name&gt;
 *   &lt;param-value&gt;X-UA-Compatible: IE=edge&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;additionalHeader.2.mimeTypeRegex&lt;/param-name&gt;
 *   &lt;param-value&gt;text/html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 * &lt;/servlet&gt;
 * </pre>
 * 
 * <p>Additional, seldom used initialization parameters are <code>minCompressionSize</code> and
 * <code>bufferSize</code>, which are actually integer values.</p>
 * 
 * <p>In OSGi frameworks it is recommended to set up this servlet using blueprint or a similar
 * technology using the supplied bean property setters like in the following example:</p>
 * <pre>
 *  &lt;bean id="resourceServlet" class="org.clazzes.util.http.ResourceServlet"&gt;
 *    &lt;property name="resourcePath" value="/foobar-resources"&gt;&lt;/property&gt;
 *    &lt;property name="excludeMimeTypes"&gt;
 *      &lt;set&gt;
 *        &lt;value&gt;image/png&lt;/value&gt;
 *        &lt;value&gt;image/gif&lt;/value&gt;
 *        &lt;value&gt;image/jpeg&lt;/value&gt;
 *        &lt;value&gt;image/tiff&lt;/value&gt;
 *      &lt;/set&gt;
 *    &lt;/property&gt;
 *    &lt;property name="aliases"&gt;
 *      &lt;map&gt;
 *        &lt;entry key="/"          value="/index.html"/&gt;       
 *        &lt;entry key="/index.htm" value="/index.html"/&gt;       
 *      &lt;/map&gt;
 *    &lt;/property&gt;
 *    &lt;property name="redirects"&gt;
 *      &lt;map&gt;
 *        &lt;entry key="/app1" value="appServlet"/&gt;       
 *        &lt;entry key="/app2" value="/ctx2/appServlet"/&gt;       
 *      &lt;/map&gt;
 *    &lt;/property&gt;
 *    &lt;property name="maxAgeSeconds" value="7200"&gt;&lt;/property&gt;
 *    &lt;property name="cacheRegex" value=".*\.cache\.\w+"&gt;&lt;/property&gt;
 *    &lt;property name="nocacheRegex" value=".*\.nocache\.\w+"&gt;&lt;/property&gt;
 *    &lt;bp:property name="additionalHeaders"&gt;
 *      &lt;bp:list&gt;
 *        &lt;bp:bean class="org.clazzes.util.http.AdditionalHeader"&gt;
 *          &lt;bp:property name="header" value="X-Frame-Options"/&gt;
 *          &lt;bp:property name="value" value="DENY"/&gt;
 *          &lt;bp:property name="mimeTypeRegex" value="text/html"/&gt;
 *        &lt;/bp:bean&gt;
 *        &lt;bp:bean class="org.clazzes.util.http.AdditionalHeader"&gt;
 *          &lt;bp:property name="header" value="X-Frame-Options"/&gt;
 *          &lt;bp:property name="value" value="SAMEORIGIN"/&gt;
 *          &lt;bp:property name="pathRegex" value=".*\.cache\.html"/&gt;
 *        &lt;/bp:bean&gt;
 *        &lt;bp:bean class="org.clazzes.util.http.AdditionalHeader"&gt;
 *          &lt;bp:property name="header" value="X-UA-Compatible"/&gt;
 *          &lt;bp:property name="value" value="IE=edge"/&gt;
 *          &lt;bp:property name="mimeTypeRegex" value="text/html"/&gt;
 *        &lt;/bp:bean&gt;
 *      &lt;/bp:list&gt;
 *    &lt;/bp:property&gt;
 *  &lt;/bean&gt;
 * </pre>
 * 
 * <p>The servlet delivers <code>Expires</code> and <code>Cache-Control</code> headers
 * according to the following rules.
 * <ol>
 * <li>
 * If the name of the resource after applying aliases
 * matches <code>cacheRegex</code>, no <code>Expires</code> and <code>Cache-Control</code>
 * headers are delivered.</li>
 * <li> If <code>maxAgeSeconds</code> is set to <code>0</code> or  the name of the resource
 * after applying aliases matches <code>nocacheRegex</code>, <code>Expires: 0</code> and
 * <code>Cache-Control: no-cache</code> headers are delivered.
 * </li>
 * <li> If <code>maxAgeSeconds</code> is set to a negative value, no <code>Expires</code> and
 * <code>Cache-Control</code> headers are delivered.
 * </li>
 * <li> Otherwise, <code>Expires</code> and
 * <code>Cache-Control: max-age=&lt;maxAgeSeconds&gt;</code>
 * are delivered according to the value of the <code>maxAgeSeconds</code> property.
 * </li>
 * </ol>
 * </p>
 * 
 * <p>In the default configuration, this servlet delivers resources by calling
 * {@link ServletContext#getResourceAsStream(String)}. If a resource class
 * loader is configured by {@link #setResourceClassLoader(ClassLoader)}
 * or {@link #setResourceClassLoaderHint(Object)}, the delivered resources
 * are acquired by calling {@link ClassLoader#getResourceAsStream(String)}
 * instead. In traditional <code>web.xml</code> setups, a resource class loader
 * may be configured by setting the init parameter
 * <code>useClassLoaderResources</code> to <code>true</code>.
 * </p>
 * 
 * <p>It is possible to registers delegate servlets as plugins using 
 * {@link #setGetPlugins(Map)}, {@link #setPostPlugins(Map)},
 * {@link #addGetPlugin(String, Servlet)} or
 * {@link #addPostPlugin(String, Servlet)} for given extensions, which are
 * applied after resolving the incoming path info using the configured aliases.
 * </p>
 */
public class ResourceServlet extends AbstractResourceSerlvlet
{
    private static final Logger log = LoggerFactory.getLogger(ResourceServlet.class);
    
    private static final long serialVersionUID = 2854439871115810811L;

    private static final DateFormat RFC822_DF = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z",Locale.US);

    private static final String ETAG_DELIM = "\""; 
    
    private String lastModified;
    private int bufferSize;
    private String resourcePath;
    private int minCompressionSize;
    private Set<String> excludeMimeTypes;
    private Map<String,String> redirects;
    private Map<String,String> aliases;
    private Map<String,LocalizedTarget> localizedRedirects;
    private Map<String,Servlet> getPlugins;
    private Map<String,Servlet> postPlugins;
    
    private ConcurrentMap<String,ResourceInfo> resourceInfos;

    private int maxAgeSeconds;
    private Pattern cacheRegex;
    private Pattern nocacheRegex;
    
    private ClassLoader resourceClassLoader;
    private boolean noETags;
    
    private static final class LocalizedTarget {
    	
    	public final String target;
    	public final String[] providedLocales;
    	
		public LocalizedTarget(String target, String[] providedLocales) {
			super();
			this.target = target;
			this.providedLocales = providedLocales;
		}
		
		public String toString() {
			return "target="+this.target+",locales="+Arrays.toString(this.providedLocales);
		}
    }
    
    private static final class ResourceInfo {
        
        public ResourceInfo(String eTag, long contentLength)
        {
            super();
            this.eTag = eTag;
            this.contentLength = contentLength;
        }
        
        public final String eTag;
        public final long contentLength;
        
        public String toString() {
            return "eTag="+this.eTag+",contentLength="+this.contentLength;
        }
    }
    
    public ResourceServlet() {
    	
    	this.minCompressionSize = 4096;
        this.bufferSize = 4096;
        this.maxAgeSeconds = 3600;
        
        this.cacheRegex = Pattern.compile(".*\\.cache\\.\\w+");
        this.nocacheRegex = Pattern.compile(".*\\.nocache\\.\\w+");
        
        this.resourceInfos = new ConcurrentHashMap<String,ResourceInfo>();
        
        this.lastModified = RFC822_DF.format(new Date());
        
        if (log.isDebugEnabled()) {
            log.debug("Servlet created, lastModified=["+this.lastModified+"].");
        }
    }
    
    
    
    private ResourceInfo enterETag(String pathInfo, InputStream is, byte[] buf) throws IOException, ServletException {
        
        int n;
        try
        {
            MessageDigest md = MessageDigest.getInstance("SHA-1");
            long l = 0;
            
            while ((n=is.read(buf)) > 0) {
            
                md.update(buf,0,n);
                l+=n;
            }
            
            Base64 b64 = new Base64(80,new byte[0]);
            String eTag = b64.encodeToString(md.digest());
            
            ResourceInfo ri = new ResourceInfo(eTag,l);
            this.resourceInfos.putIfAbsent(pathInfo,ri);
            return ri;

        } catch (NoSuchAlgorithmException e)
        {
            throw new ServletException("SHA-1 digest not found",e);
        }
       
    }
    
    /**
     * This method returns the result of
     * {@link ServletContext#getResourceAsStream(String)} if the
     * resource class loader is <code>null</code> or otherwise
     * calls <code>resourceClassLoader.getResourceAsStream()</code>.
     * 
     * @param resPath The final resource path to open.
     * @return An input stream to read from.
     */
    protected InputStream getResourceAsStream(String resPath) {
    	
    	if (this.resourceClassLoader == null) {
    	
    		return this.getServletContext().getResourceAsStream(resPath);
    	}
    	else {
    		return this.resourceClassLoader.getResourceAsStream(resPath);
    	}
    }
    
    /**
     * Clear all cached checksums for delivering ETags. This method
     * might be needed by classes overriding {@link #getResourceAsStream(String)}.
     */
    protected void clearResourceInfos() {
    	this.resourceInfos.clear();
    }
    
    /**
     * Resolve a path using the configured list of aliases.
     * 
     * @param path The path, which should be set to the empty string, if
     *             {@link HttpServletRequest#getPathInfo()} returned
     *             <code>null</code>.
     * @return The path mapped to it's alias or the original path, if no alias
     *             is configured.
     */
    protected String resolvePath(String path) {
    	
        if (this.aliases != null) {
            
            String alias = this.aliases.get(path);
            
            if (alias != null) {
             
                if (log.isDebugEnabled()) {
                    log.debug("Resolved ["+path+"] to alias ["+alias+"].");
                }
                
                return alias;
            }
        }
        
        return path;
    }
    
    /**
     * Call a plugin based on the extension of the resolved path.
     * 
     * @param path The original path, which is the empty string ,if {@link HttpServletRequest#getPathInfo()} 
     *             returned <code>null</code>.
     * @param resolvedPath The resolved server-side path to be propagated via
     *          {@link RequestHelper#setResolvedPath(HttpServletRequest, String)} if
     *          it differs from <code>path</code>
     * @param plugins A list of plugins by their extensions. 
     * @param req The servlet request.
     * @param resp The serlvet response.
     * @return Whether a plugin has been found and called.
     * @throws ServletException Upon failure calling the plugin.
     * @throws IOException Upon failure calling the plugin.
     */
    protected static boolean maybeCallPlugin(String path,
    		String resolvedPath, Map<String,Servlet> plugins,
    		HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
    	
    	Servlet plugin = null;
        
    	int ep = resolvedPath.lastIndexOf('.');
        
    	if (ep >= 0) {
    	
    		int sp = resolvedPath.indexOf('/',ep);
    		
    		if (sp < 0) {
    			
    			String ext = resolvedPath.substring(ep+1);
    			
    			plugin = plugins.get(ext);
    			
    			if (log.isDebugEnabled()) {
    				
    				if (plugin == null) {

    					log.debug("No plugin found for extension [{}] of resolved path [{}].",ext,resolvedPath);
    				}
    				else {
    					log.debug("Calling plugin [{}] for resolved path [{}].",plugin.getServletInfo(),resolvedPath);
    				}
    			}
    		}
    	}

    	if (plugin == null) {
    		return false;
    	}
    	
    	if (resolvedPath != path) {
    		RequestHelper.setResolvedPath(req,resolvedPath);
    	}
    	
    	plugin.service(req,resp);
    	return true;
    }
    
    /* (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException,
                                                                          IOException
    {
        String path = req.getPathInfo();
        
        // some servlet container send a null path info, if the servlet path
        // itself is specified without any extra path and without a slash.
        // replace null with empty string in order to allow easier configuration
        // of redirects and aliases for the root context.
        if (path == null) path="";
        
        if (log.isDebugEnabled()) {
            log.debug("Received GET request ["+req.getServletPath()+path+"]");
        }

        Date now = new Date();
        
        if (this.localizedRedirects != null) {
        	
            LocalizedTarget lt = this.localizedRedirects.get(path);
            
         	if (lt != null) {
            
         		String acceptLanguage = req.getHeader("Accept-Language");
                
         		if (acceptLanguage != null) {
         			String loc = LocaleHelper.selectLocale(acceptLanguage,lt.providedLocales);
        		
         			if (loc != null) {
         				
         				String redirect = lt.target + loc;
         	             
                        if (log.isDebugEnabled()) {
                            log.debug("Redirecting request ["+path+"] to localized target ["+redirect+"].");
                        }
                        
                        resp.setHeader("Date",RFC822_DF.format(now));
                        resp.sendRedirect(redirect);
                        return;
        			}
         		}
        	}
        }
        
        
        if (this.redirects != null) {
            
            String redirect = this.redirects.get(path);
            
            if (redirect != null) {
             
                if (log.isDebugEnabled()) {
                    log.debug("Redirecting request ["+path+"] to ["+redirect+"].");
                }
                
                resp.setHeader("Date",RFC822_DF.format(now));
                resp.sendRedirect(redirect);
                return;
            }
        }
        
        String resolvedPath = this.resolvePath(path);
        
        if (this.getPlugins != null) {
        	
        	if (maybeCallPlugin(path, resolvedPath, this.getPlugins, req, resp)) {
        		return;
        	}
        }
        
        if (this.resourcePath == null) {
        	
            resp.setHeader("Date",RFC822_DF.format(now));
            resp.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }
        
        InputStream is = null;
        OutputStream os = null;
        
        try {
            
            String res = this.resourcePath + resolvedPath;
            
            if (log.isDebugEnabled()) {
                log.debug("Resolving resource ["+res+"]");
            }
            
            is = this.getResourceAsStream(res);

            if (is == null) {
                
                log.error("Resource ["+res+"] not found.");
                resp.sendError(HttpServletResponse.SC_NOT_FOUND);
            }
            else {
                
            	String mimeType = this.getServletContext().getMimeType(resolvedPath);
                ResourceInfo ri = null;
            	boolean changed = true;
        		byte[] buf = null;
        		            	
            	if (!this.noETags) {
            		String inm = req.getHeader("If-None-Match");
            		String[] clientETags = inm == null ? null : inm.split(",");
            		
            		if (log.isDebugEnabled()) {
            			log.debug("Client sent ETags ["+Arrays.toString(clientETags)+"]");
            		}
            		
            		ri = this.resourceInfos.get(resolvedPath);
            		
            		if (ri == null) {
            			
            			if (log.isDebugEnabled()) {
            				log.debug("Calculating ResourceInfo for resource ["+resolvedPath+"]");
            			}
            			
            			buf = new byte[this.bufferSize];
            			ri = this.enterETag(resolvedPath,is,buf);
            			is.close();
            			is = this.getResourceAsStream(res);
            		}
            		
            		if (log.isDebugEnabled()) {
            		    log.debug("Got ResourceInfo ["+ri+"]");
            		}
            		
            		if (clientETags != null) {
            			for (String clientETag : clientETags) {
            				
            				String ce = clientETag.trim();
            				
            				if (ce.startsWith(ETAG_DELIM) && ce.endsWith(ETAG_DELIM) &&
            						ce.substring(1,ce.length()-1).equals(ri.eTag)) {
            					changed = false;
                            break;
            				}
            			}
            		}
            	}
            	
            	this.propagateAdditionalHeaders(resp,resolvedPath,mimeType);
            	
                if (ri != null) {
                	resp.setHeader("ETag",ETAG_DELIM + ri.eTag + ETAG_DELIM);
                }
                
                if (mimeType != null)
                    resp.setContentType(mimeType);
                
                if (changed) {
                    
                	boolean cacheable =
                			this.cacheRegex != null &&
                			this.cacheRegex.matcher(resolvedPath).matches();

                	if (!cacheable) {
                	
                		int maxAgeSecs = this.getMaxAgeSeconds();
                		
                		boolean nocacheable = maxAgeSecs == 0 ||
                				(this.nocacheRegex != null && this.nocacheRegex.matcher(resolvedPath).matches());

                		if (nocacheable) {
                			ResponseHelper.setNoCacheHeaders(resp);
                		}
                		else if (maxAgeSecs > 0) {
 
                			resp.setHeader("Expires",RFC822_DF.format(new Date(now.getTime()+1000L*maxAgeSecs)));
                			resp.setHeader("Cache-Control","max-age="+maxAgeSecs);
                		}
                	}
                    
                	// Send a last modified header only, if noETags is set to false.
                	// noETags is mostly set in development environments, where the delivered
                	// resources are expected to change during the lifetime of the servlet
                	// instance.
                	// Hence, it is not feasible to send a Last-Modified header in these situations.
                	if (ri != null) {
                		resp.setHeader("Last-Modified",this.lastModified);
                	}
                	
                    if (buf == null)
                        buf = new byte[this.bufferSize];
                    
                    boolean compress = (ri == null || ri.contentLength > this.minCompressionSize) &&
                    (this.excludeMimeTypes == null || !this.excludeMimeTypes.contains(mimeType));
                    
                    if (compress) {
                        os = CompressionHelper.setupResponseOutputStream(req,resp); 
                        
                        if (log.isDebugEnabled()) {
                            log.debug("compressed=["+CompressionHelper.isCompressed(os)+"].");
                        }
                    }
                    else {
                        
                        if (log.isDebugEnabled()) {
                            log.debug("Compression dissabled by configuration.");
                        }
                        
                        os = resp.getOutputStream();
                    }
                    
                    if (ri != null && (!compress || !CompressionHelper.isCompressed(os))) {
                        // don't use resp.setContentLength(int), because it's restricted
                        // to 31 bits aka 2 GB of data.
                        resp.setHeader("Content-Length",Long.toString(ri.contentLength));
                    }
                    
                    int n;
                    
                    while ((n=is.read(buf)) > 0) {
                        
                        os.write(buf,0,n);
                    }

                    if (log.isDebugEnabled()) {
                        log.debug("Finished request ["+path+"]");
                    }
                }
                else {
                    if (log.isDebugEnabled()) {
                        log.debug("Sending "+HttpServletResponse.SC_NOT_MODIFIED+" not modified.");
                    }
                    
                    resp.sendError(HttpServletResponse.SC_NOT_MODIFIED);
                }
            }
        }
        finally {
            if (os != null) CompressionHelper.finishOutputStream(os);
            if (is != null) is.close();
        }
    }

    @Override
	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {

        if (this.postPlugins != null) {
        	
        	String path = req.getPathInfo();
        	
        	// some servlet container send a null path info, if the servlet path
        	// itself is specified without any extra path and without a slash.
        	// replace null with empty string in order to allow easier configuration
        	// of redirects and aliases for the root context.
        	if (path == null) path="";
        
        	if (log.isDebugEnabled()) {
        		log.debug("Received POST request ["+req.getServletPath()+path+"]");
        	}
        	
        	String resolvedPath = this.resolvePath(path);
        	
        	if (maybeCallPlugin(path, resolvedPath, this.postPlugins, req, resp)) {
        		return;
        	}
        }
        
        // send a 405 error as per servlet APIs default implementation.
		super.doPost(req, resp);
	}

	/**
     * @return the internal buffer size.
     */
    public int getBufferSize()
    {
        return this.bufferSize;
    }

    /**
     * @param bufferSize the internal buffer size to set. The default value is 4096 and should be
     *                   suitable for almost any purpose.
     */
    public void setBufferSize(int bufferSize)
    {
        this.bufferSize = bufferSize;
    }

    /**
     * @return the resource path, which is exported from the given ClassLoader. If set to
     *           <code>null</code>, only redirections are available.
     */
    public String getResourcePath()
    {
        return this.resourcePath;
    }

    /**
     * @param resourcePath the resource path, which is exported from the given ClassLoader to set.
     *           All request exports resources acquired through
     *           {@link ServletContext#getResourceAsStream(String)} using this resource path
     *           appended by {@link HttpServletRequest#getPathInfo()}.
     */
    public void setResourcePath(String resourcePath)
    {
        this.resourcePath = resourcePath;
    }

    /**
     * @return the the minimal size of objects to be compressed.
     */
    public int getMinCompressionSize()
    {
        return this.minCompressionSize;
    }

    /**
     * @param minCompressionSize the minimal size of objects to be compressed to set.
     */
    public void setMinCompressionSize(int minCompressionSize)
    {
        this.minCompressionSize = minCompressionSize;
    }

    /**
     * @return The list of MIME types to be excluded from compression.
     */
    public Set<String> getExcludeMimeTypes()
    {
        return this.excludeMimeTypes;
    }

    /**
     * @param excludeMimeTypes The list of MIME types to be excluded from compression to set.
     *               Most image types are already compressed, so should consider to include
     *               their MIME types in this list.
     */
    public void setExcludeMimeTypes(Set<String> excludeMimeTypes)
    {
        this.excludeMimeTypes = excludeMimeTypes;
    }

    /**
     * Add an element to the list of MIME types to be excluded from compression.
     * @param excludeMimeType The MIME type to add.
     */
    public void addExcludeMimeType(String excludeMimeType) {
        
        if (this.excludeMimeTypes == null)
            this.excludeMimeTypes = new HashSet<String>();
        
        this.excludeMimeTypes.add(excludeMimeType);
    }
    
    /* (non-Javadoc)
     * @see javax.servlet.GenericServlet#getServletInfo()
     */
    @Override
    public String getServletInfo()
    {
        return "ResourceServlet[" + this.resourcePath +"]";
    }

    private static Map<String,String> parseMapping(String mapping) throws ServletException {
        
        String[] emt = mapping.split(",");
  
        Map<String,String> emts = new HashMap<String,String>();
        
        for (String mt : emt) {
            
            String[] x = mt.trim().split("=",2);
            
            if (x.length < 2)
                throw new ServletException("Invalid path mapping part ["+mt+"] specified (should contain an '=' character).");
                
            emts.put(x[0],x[1]);
        }
        return emts;
    }
    
    /* (non-Javadoc)
     * @see javax.servlet.GenericServlet#init(javax.servlet.ServletConfig)
     */
    @Override
    public void init() throws ServletException
    {
        super.init();
        
        ServletConfig cfg = this.getServletConfig();
        
        String p = cfg.getInitParameter("resourcePath");
        
        if (p != null)
            this.setResourcePath(p);
        
        p = cfg.getInitParameter("excludeMimeTypes");
        
        if (p != null) {
            
            String[] emt = p.split(",");
            Set<String> emts = new HashSet<String>();
            for (String mt : emt) {
                emts.add(mt.trim());
            }
            this.setExcludeMimeTypes(emts);
        }
        
        p = cfg.getInitParameter("bufferSize");
        
        if (p != null) {
            this.setBufferSize(Integer.parseInt(p));
        }
        
        p = cfg.getInitParameter("maxAgeSeconds");
        
        if (p != null) {
            this.setMaxAgeSeconds(Integer.parseInt(p));
        }
        
        p = cfg.getInitParameter("cacheRegex");
        
        if (p != null) {
            this.setCacheRegex(Pattern.compile(p));
        }
        
        p = cfg.getInitParameter("nocacheRegex");
        
        if (p != null) {
            this.setNocacheRegex(Pattern.compile(p));
        }
        
        p = cfg.getInitParameter("minCompressionSize");
        
        if (p != null) {
            this.setMinCompressionSize(Integer.parseInt(p));
        }
        
        p = cfg.getInitParameter("aliases");
        
        if (p != null) {
            this.setAliases(parseMapping(p));
        }
        
        p = cfg.getInitParameter("redirects");
        
        if (p != null) {
            this.setRedirects(parseMapping(p));
        }
        
        p = cfg.getInitParameter("noETags");
        
        if (p != null) {
            this.setNoETags(Boolean.parseBoolean(p));
        }
        
        p = cfg.getInitParameter("useClassLoaderResources");
        
        if (p != null) {
        	if (Boolean.parseBoolean(p)) {
        		this.setResourceClassLoader(ResourceServlet.class.getClassLoader());
        	}
        	else {
        		this.setResourceClassLoader(null);
        	}
        }
    }

    /**
     * @param seconds The value of the <code>Cache-Control: max-age=&lt;sec&gt;</code>
     *         header delivered to all resources, which are not explicitly marked
     *         as cacheable by matching against {@link #getCacheRegex()}.
     *         If set to <code>0</code> all resources are marked as non-cacheable.
     *         If set to <code>-1</code> all resources are delivered without
     *         a Cache-Control header.
     */
    public synchronized void setMaxAgeSeconds(int seconds) {
		this.maxAgeSeconds = seconds;
	}

    /**
     * @return The value of the <code>Cache-Control: max-age=&lt;sec&gt;</code>
     *         header delivered to all resources, which are not explicitly marked
     *         as cacheable by matching against {@link #getCacheRegex()}.
     */
    public synchronized int getMaxAgeSeconds() {
    	return this.maxAgeSeconds;
    }

	/**
	 * @return A regular expression, which is used to determine resources, which
	 *         are delivered with no <code>Expires:</code> and
	 *         <code>Cache-Control: max-age=&lt;max-age&gt;</<code>  headers.
	 *         The default value is <code>".*\.cache\.\w+"</code>
	 */
	public Pattern getCacheRegex() {
		return this.cacheRegex;
	}

	/**
	 * @param cacheRegex A regular expression, which is used to determine resources, which
	 *                   are delivered with no <code>Expires:</code> and
	 *                   <code>Cache-Control: max-age=&lt;max-age&gt;</<code>  headers.
	 *                   The default value is <code>".*\.cache\.\w+"</code>
	 */
	public void setCacheRegex(Pattern cacheRegex) {
		this.cacheRegex = cacheRegex;
	}

	/**
	 * @return A regular expression, which is used to determine resource,
	 *         which are delivered with <code>Expires: 0</code> and
	 *         <code>Cache-Control: no-cache</code>  headers.
	 *         The default value is <code>".*\.nocache\.\w+"</code>
	 */
	public Pattern getNocacheRegex() {
		return this.nocacheRegex;
	}

	/**
	 * @param nocacheRegex A regular expression, which is used to determine resource,
	 *               which are delivered with <code>Expires: 0</code> and
	 *               <code>Cache-Control: no-cache</code>  headers.
	 *               The default value is <code>".*\.nocache\.\w+"</code>
	 */
	public void setNocacheRegex(Pattern nocacheRegex) {
		this.nocacheRegex = nocacheRegex;
	}

	/**
     * @return The mapping of requests path infos to redirect URLs.
     */
    public Map<String, String> getRedirects()
    {
        return this.redirects;
    }

    private void inspectRedirectForLocalization(String key, String value) {
    	
       	int idx = key.indexOf('?');
		
    	if (idx >= 0) {
    		
    		String q = key.substring(idx+1);
    		
    		if (q.startsWith("locales=")) {
    			
    			String[] providedLocales = q.substring(8).split(",");
    			String target;
    			
    			if (value.indexOf('?') >= 0) {
    				target = value + "&locale=";
    			}
    			else {
    				target = value + "?locale=";        				
    			}
    			
    			if (this.localizedRedirects == null)
    				this.localizedRedirects = new HashMap<String,LocalizedTarget>();

    			String barekey = key.substring(0,idx);
    			LocalizedTarget lt = new LocalizedTarget(target, providedLocales);
    			
    			if (log.isDebugEnabled()) {
    				log.debug("Adding localized redirect from [{}] to [{}].",
    						barekey,lt);
    			}
    			
    			this.localizedRedirects.put(barekey,lt);
    		}
    	}
    }
    
    /**
     * <p>The keys of this map are path infos like returned by
     * {@link HttpServletRequest#getPathInfo()}, which starts with a '/' sign.
     * The values are relative or absolute URLs, which are passed to
     * {@link HttpServletResponse#sendRedirect(String)}.</p>
     * 
     * <p>The key may be optionally postfixed by a list of locales in the format
     * <code>/localized-url?locales=de,en</code> meaning that any request for
     * <code>/localized-url</code> with an appropriate <code>Accept-Language</code>
     * HTTP header will be redirected to the target with an URL parameter named
     * <code>locale</code> with a value of the selected domain as per
     * {@link LocaleHelper#selectLocale(String, String[])} using the comma-separated
     * list of server-side locales <code>de</code> or <code>en</code>.
     *  </p>
     *
     * @param redirects The mapping of requests path infos to redirect URLs to set.
     */
    public void setRedirects(Map<String, String> redirects)
    {
        this.redirects = redirects;
        
        // scan for localized redirects.
        this.localizedRedirects = null;
        
        for (Map.Entry<String,String> e:redirects.entrySet()) {
        
        	this.inspectRedirectForLocalization(e.getKey(),e.getValue());
        }
    }

    /**
     * Add a path-to-URL redirect.
     * 
     * @param pathInfo A path info as returned by {@link HttpServletRequest#getPathInfo()}.
     * @param url A redirect URL passed to {@link HttpServletResponse#sendRedirect(String)}.
     */
    public void addRedirect(String pathInfo, String url) {
        
        if (this.redirects == null)
            this.redirects= new HashMap<String, String>();
        
        this.redirects.put(pathInfo,url);
    	this.inspectRedirectForLocalization(pathInfo,url);
    }
    
    /**
     * @return the resource aliases.
     */
    public Map<String, String> getAliases()
    {
        return this.aliases;
    }

    /**
     * @param aliases the resource aliases to set.
     *         The keys of this map are path infos like returned by
     *         {@link HttpServletRequest#getPathInfo()}, which start with a '/' sign.
     *         The values are translated path infos and should therefore start with
     *         a '/' sign, too.
     */
    public void setAliases(Map<String, String> aliases)
    {
        this.aliases = aliases;
    }
    
    /**
     * Add an alias for a given path info.
     * 
     * @param pathInfo A path info as returned by {@link HttpServletRequest#getPathInfo()}.
     * @param alias An alias path info, which should start with a '/' sign.
     */
    public void addAlias(String pathInfo, String alias) {
        
        if (this.aliases == null)
            this.aliases = new HashMap<String, String>();
        
        this.aliases.put(pathInfo,alias);
    }

	/**
	 * @return a class loader to resolve resources from instead of
	 *         resolving from {@link ServletContext#getResourceAsStream(String)}
	 */
	public ClassLoader getResourceClassLoader() {
		return this.resourceClassLoader;
	}

	/**
	 * @param resourceClassLoader a class loader to resolve resources
	 *         from instead of resolving from
	 *         {@link ServletContext#getResourceAsStream(String)} to set.
	 */
	public void setResourceClassLoader(ClassLoader resourceClassLoader) {
		this.resourceClassLoader = resourceClassLoader;
	}

	/**
	 * Set the resource class loader according to the given hint object.
	 * If <code>hint</code> is <code>null</code>, the resource class
	 * loader is set to <code>null</code>. If hint is an instance of
	 * {@link ClassLoader}, the resource class loader is set to
	 * <code>hint</code> .If hint is an instance of
	 * {@link Class}, the resource class loader is set to
	 * <code>hint.getClassLoader()</code>. Otherwise, the resource
	 * class loader is set to
	 * <code>hint.getClass().getClassLoader()</code>.
	 * 
	 * @param hint An object to deduce the resource class loader from.
	 */
	public void setResourceClassLoaderHint(Object hint) {
		
		if (hint == null) {
			this.resourceClassLoader = null;
		}
		else if (hint instanceof ClassLoader) {
			this.resourceClassLoader = (ClassLoader)hint;
		}
		else if (hint instanceof Class) {
			this.resourceClassLoader = ((Class<?>)hint).getClassLoader();
		}
		else {
			this.resourceClassLoader = hint.getClass().getClassLoader();
		}
	}

	/**
	 * @return Whether no ETag-specific headers should be sent.
	 *         If set to <code>true</code>, the <code>Last-Modified</code>
	 *         header will also be omitted.
	 */
	public boolean isNoETags() {
		return this.noETags;
	}

	/**
	 * @param noETags Whether no ETag-specific headers should be sent.
	 *         If set to <code>true</code>, the <code>Last-Modified</code>
	 *         header will also be omitted.
	 */
	public void setNoETags(boolean noETags) {
		this.noETags = noETags;
	}

	/**
	 * @return The map of registered plugins for GET requests by
	 *         their extension.
	 */
	public Map<String, Servlet> getGetPlugins() {
		return this.getPlugins;
	}

	/**
	 * Add a plugin for GET requests for a given extension.
	 * 
	 * @param extension The extension, which will be matched against the resolved path.
	 * @param plugin The serlvet to call for requests on resources with the given extension.
	 */
	public void addGetPlugin(String extension, Servlet plugin) {
		
		if (this.getPlugins == null) {
			this.getPlugins = new HashMap<String,Servlet>();
		}
		
		this.getPlugins.put(extension,plugin);
	}
	
	/**
	 * @param getPlugins The map of registered plugins for GET requests by
	 *         their extension to set.
	 */
	public void setGetPlugins(Map<String, Servlet> getPlugins) {
		this.getPlugins = getPlugins;
	}

	/**
	 * @return The map of registered plugins for POST requests by
	 *         their extension.
	 */
	public Map<String, Servlet> getPostPlugins() {
		return this.postPlugins;
	}

	/**
	 * Add a plugin for POST requests for a given extension.
	 * 
	 * @param extension The extension, which will be matched against the resolved path.
	 * @param plugin The serlvet to call for requests on resources with the given extension.
	 */
	public void addPostPlugin(String extension, Servlet plugin) {
		
		if (this.postPlugins == null) {
			this.postPlugins = new HashMap<String,Servlet>();
		}
		
		this.postPlugins.put(extension,plugin);
	}

	/**
	 * @param postPlugins The map of registered plugins for POST requests by
	 *         their extension to set.
	 */
	public void setPostPlugins(Map<String, Servlet> postPlugins) {
		this.postPlugins = postPlugins;
	}

}
